class BigInteger:
    pass


class EmailString:
    pass


class HostnameString:
    pass


class IP4String:
    pass


class NullOrBoolean:
    pass


class PositiveDecimal:
    pass


class PositiveInteger:
    pass


class PositiveSmallInteger:
    pass


class SmallInteger:
    pass


class Text:
    pass
